import asyncio
import importlib
from typing import final
from uuid import uuid4

from cow.monitoring.manager import LunaMonitoringManager
from cow.web.application import BaseLunaApplicationCtx, LunaApplication as _LunaApplication
from luna3.client import Luna3Session
from luna3.settings import HTTPSettings
from sanic import HTTPResponse

from .configs.configs.services import SettingsLambdaTasksUnit, SettingsLambdaUnit
from .configs.configs.settings.classes import ServiceAddressSettings, ServiceTimeoutsSettings
from .monitoring import MonitoringPoint
from .request import BaseHandlersRequest, BaseLambdaRequest, BaseStandaloneLambdaRequest
from .tasks_tools.base_subtask import BaseSubTask, SubTaskToImageStore
from .tasks_tools.monitoring.points import SubtaskMonitoringBasePoint, WorkerRequestPoint
from .tasks_tools.redis_db.context import WorkerRedisContext


class FakeManager:
    """Fake plugin manager"""

    @classmethod
    async def shutdown(cls):
        """Prevent default luna-application behaviour"""

    @staticmethod
    async def sendEventToPlugins(*args, **kwargs):
        """Prevent default luna-application behaviour"""


class CtxClass(BaseLunaApplicationCtx):
    """Empty context class for lambda application"""

    pluginManager = FakeManager
    lambdaVersion: int = 1
    luna3Session: Luna3Session | None
    userCtx: object | None = None
    lambdaName: str | None = None


async def lambdaVersionMiddleware(request: BaseHandlersRequest | BaseStandaloneLambdaRequest, response: HTTPResponse):
    """
    Monitoring middleware. Generate a request point and error point and send them to app monitoring.

    Args:
        request: request
        response: response
    """
    response.headers["Lambda-Version"] = request.app.ctx.lambdaVersion


def getServiceClientKwargs(addressSetting: ServiceAddressSettings, timeoutSettings: ServiceTimeoutsSettings) -> dict:
    """Get service client kwargs"""
    return {
        "origin": addressSetting.origin,
        "api": addressSetting.apiVersion,
        "totalTimeout": timeoutSettings.totalTimeout,
        "connectTimeout": timeoutSettings.connectTimeout,
        "sockConnectTimeout": timeoutSettings.sockConnectTimeout,
        "sockReadTimeout": timeoutSettings.sockReadTimeout,
        "asyncRequest": True,
    }


class BaseLambdaApp(_LunaApplication):
    """Base for lambda application"""

    @final
    def __init__(self, serviceName: str, lambdaVersion: int, lambdaType: str, userCtxCls: type[object]):
        CtxClass.lambdaVersion = lambdaVersion
        from ..public.handlers.handlers_request import HandlersLambdaRequest
        from ..public.standalone.standalone_request import StandaloneLambdaRequest

        super().__init__(
            apiVersion=1,
            configClass=SettingsLambdaUnit,
            defaultPort=8000,
            ctxClass=CtxClass,
            requestClass=HandlersLambdaRequest if lambdaType == "handlers" else StandaloneLambdaRequest,
        )
        self.response_middleware.append(lambdaVersionMiddleware)
        self.ctx.luna3Session = None
        self.ctx.lambdaType = lambdaType
        self.ctx.userCtx = userCtxCls()
        # hack to override lambda name for monitoring
        self.ctx.lambdaName = serviceName

    @final
    async def shutdown(self):
        if hasattr(self.ctx.userCtx, "onShutdown"):
            return await self.ctx.userCtx.onShutdown()

    @final
    async def initialize(self):
        """Description see :func:`~_LunaApplication.initialize`"""
        # it is not supposed to use plugins by lambda unit
        self.ctx.serviceConfig.activePlugins = []
        await super().initialize()
        self.ctx.luna3Session = Luna3Session(**self.getLuna3Kwargs(self.ctx.serviceConfig))
        if hasattr(self.ctx.userCtx, "onStart"):
            return await self.ctx.userCtx.onStart()
        MonitoringPoint.initialize(serviceName=self.ctx.lambdaName)

    @staticmethod
    def getLuna3Kwargs(config: SettingsLambdaUnit):
        lunaKwargs = {
            "facesSettings": HTTPSettings(
                **getServiceClientKwargs(addressSetting=config.facesAddress, timeoutSettings=config.facesTimeouts),
            ),
            "remoteSDKSettings": HTTPSettings(
                **getServiceClientKwargs(
                    addressSetting=config.remoteSDKAddress, timeoutSettings=config.remoteSDKTimeouts
                ),
            ),
        }
        if config.additionalServicesUsage.lunaImageStore:
            lunaKwargs.update(
                {
                    "faceSampleStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.faceSamplesStorage, timeoutSettings=config.faceSamplesStorageTimeouts
                        ),
                    ),
                    "bodySampleStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.bodySamplesStorage, timeoutSettings=config.bodySamplesStorageTimeouts
                        ),
                    ),
                    "imageOriginStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.imageOriginStorage, timeoutSettings=config.imageOriginStorageTimeouts
                        ),
                    ),
                }
            )
        if config.additionalServicesUsage.lunaEvents:
            lunaKwargs["eventSettings"] = HTTPSettings(
                **getServiceClientKwargs(addressSetting=config.eventsAddress, timeoutSettings=config.eventsTimeouts),
            )
        if config.additionalServicesUsage.lunaMatcherProxy:
            lpmAddress, lpmTimeouts = config.matcherProxyAddress, config.matcherProxyTimeouts
        else:
            lpmAddress, lpmTimeouts = config.pythonMatcherAddress, config.pythonMatcherTimeouts
        lunaKwargs["pythonMatcherSettings"] = HTTPSettings(
            **getServiceClientKwargs(addressSetting=lpmAddress, timeoutSettings=lpmTimeouts)
        )
        if config.additionalServicesUsage.lunaSender:
            lunaKwargs["senderSettings"] = HTTPSettings(
                origin=config.senderAddress.origin,
                api=config.senderAddress.apiVersion,
                asyncRequest=True,
            )
        return lunaKwargs


class LambdaTasksCtx(BaseLunaApplicationCtx):
    """Tasks worker application context"""

    #: luna 3 session
    luna3Session: Luna3Session | None
    #: worker redis db context
    workerRedisDBContext: WorkerRedisContext | None
    #: worker tasks
    tasks: dict[BaseSubTask, asyncio.Task]
    #: worker separation tasks
    separationTasks: dict[BaseSubTask, asyncio.Task]
    #: lambda task version
    lambdaVersion: int = 1
    #: user ctx
    userCtx: object | None = None
    #: lambda task name
    lambdaName: str | None = None


class LambdaTasksApp(_LunaApplication):
    """Base for lambda tasks application"""

    @final
    def __init__(self, serviceName: str, lambdaVersion: int, lambdaType: str, userCtxCls: type[object]):
        LambdaTasksCtx.lambdaVersion = lambdaVersion
        super().__init__(
            apiVersion=1,
            configClass=SettingsLambdaTasksUnit,
            defaultPort=8000,
            ctxClass=LambdaTasksCtx,
            requestClass=BaseLambdaRequest,
        )
        self.response_middleware.append(lambdaVersionMiddleware)
        self.ctx.luna3Session = None
        self.ctx.lambdaType = lambdaType
        self.ctx.userCtx = userCtxCls()
        lambdaName = serviceName
        if lambdaName.startswith("lambda-"):
            lambdaName = lambdaName[len("lambda-") :]
        self.ctx.lambdaName = lambdaName
        self.ctx.tasks = {}
        self.ctx.separationTasks = {}

    @final
    async def shutdown(self):
        if hasattr(self.ctx.userCtx, "onShutdown"):
            return await self.ctx.userCtx.onShutdown()

        await self.ctx.workerRedisDBContext.close()

        for task, asyncioTask in self.ctx.tasks.items():
            await task.onShutDown()
            asyncioTask.cancel()
        for asyncioTask in self.ctx.separationTasks.values():
            asyncioTask.cancel()

    @final
    async def initialize(self):
        """Description see :func:`~_LunaApplication.initialize`"""
        # it is not supposed to use plugins by lambda unit
        self.ctx.serviceConfig.activePlugins = []
        await super().initialize()
        self.ctx.luna3Session = Luna3Session(**self.getLuna3Kwargs(self.ctx.serviceConfig))
        if hasattr(self.ctx.userCtx, "onStart"):
            return await self.ctx.userCtx.onStart()
        MonitoringPoint.initialize(serviceName=self.ctx.lambdaName)

        if self.ctx.serviceConfig.additionalServicesUsage.lunaImageStore:
            self.initializeSubTaskToImageStore(settings=self.ctx.serviceConfig)

        self.initializeBaseWorkerClass(settings=self.ctx.serviceConfig, monitoring=self.ctx.monitoring)
        self.initializeMonitoringPoints(serviceName=self.ctx.lambdaName)
        self.ctx.workerRedisDBContext = WorkerRedisContext(self.ctx.serviceConfig.redisDB)

        kwargs = {
            "redisContext": self.ctx.workerRedisDBContext,
            "luna3Client": self.ctx.luna3Session.getClient(str(uuid4())),
            "app": self,
        }
        try:
            lambdaTaskClass = importlib.import_module("lambda_main").LambdaTask
            tsk = lambdaTaskClass(**kwargs)
            self.ctx.tasks[tsk] = asyncio.create_task(tsk.start(lambdaId=self.ctx.lambdaName))
            self.ctx.separationTasks[tsk] = asyncio.create_task(tsk.separationStart(lambdaId=self.ctx.lambdaName))
        except Exception as exc:
            excArgs = " ".join(getattr(exc, "args", []))
            raise RuntimeError(f"Lambda tasks initialize error: {str(exc)} {excArgs}")

    @staticmethod
    def getLuna3Kwargs(config: SettingsLambdaTasksUnit):
        lunaKwargs = {
            "facesSettings": HTTPSettings(
                **getServiceClientKwargs(addressSetting=config.facesAddress, timeoutSettings=config.facesTimeouts),
            ),
            "remoteSDKSettings": HTTPSettings(
                **getServiceClientKwargs(
                    addressSetting=config.remoteSDKAddress, timeoutSettings=config.remoteSDKTimeouts
                ),
            ),
        }
        if config.additionalServicesUsage.lunaImageStore:
            lunaKwargs.update(
                {
                    "faceSampleStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.faceSamplesStorage, timeoutSettings=config.faceSamplesStorageTimeouts
                        ),
                    ),
                    "bodySampleStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.bodySamplesStorage, timeoutSettings=config.bodySamplesStorageTimeouts
                        ),
                    ),
                    "imageOriginStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.imageOriginStorage, timeoutSettings=config.imageOriginStorageTimeouts
                        ),
                    ),
                    "tasksResultStoreSettings": HTTPSettings(
                        **getServiceClientKwargs(
                            addressSetting=config.taskResultStorage, timeoutSettings=config.taskResultStorageTimeouts
                        ),
                    ),
                }
            )
        if config.additionalServicesUsage.lunaHandlers:
            lunaKwargs["handlersSettings"] = HTTPSettings(
                **getServiceClientKwargs(addressSetting=config.handlersAddress, timeoutSettings=config.handlersTimeouts)
            )
        if config.additionalServicesUsage.lunaMatcherProxy:
            lpmAddress, lpmTimeouts = config.matcherProxyAddress, config.tasksToMatcherTimeouts
        else:
            lpmAddress, lpmTimeouts = config.pythonMatcherAddress, config.tasksToMatcherTimeouts
        lunaKwargs["pythonMatcherSettings"] = HTTPSettings(
            **getServiceClientKwargs(addressSetting=lpmAddress, timeoutSettings=lpmTimeouts)
        )
        if config.additionalServicesUsage.lunaEvents:
            lunaKwargs["eventSettings"] = HTTPSettings(
                **getServiceClientKwargs(addressSetting=config.eventsAddress, timeoutSettings=config.eventsTimeouts),
            )
        if config.additionalServicesUsage.lunaSender:
            lunaKwargs["senderSettings"] = HTTPSettings(
                api=config.senderAddress.apiVersion, origin=config.senderAddress.origin
            )
        return lunaKwargs

    @staticmethod
    def initializeSubTaskToImageStore(settings: SettingsLambdaTasksUnit):
        """
        Initialize SubTask to image store class.
        Args:
            settings: task worker settings
        """
        SubTaskToImageStore.initFromConfig(
            resultStoreBucket=settings.taskResultStorage.bucket,
            faceSampleBucket=settings.faceSamplesStorage.bucket,
            bodySampleBucket=settings.bodySamplesStorage.bucket,
            imageOriginBucket=settings.imageOriginStorage.bucket,
            taskResultStorageTimeouts=settings.taskResultStorageTimeouts,
            faceSamplesStorageTimeouts=settings.faceSamplesStorageTimeouts,
            bodySamplesStorageTimeouts=settings.bodySamplesStorageTimeouts,
            imageOriginStorageTimeouts=settings.imageOriginStorageTimeouts,
            imageStoreRequestConcurrency=settings.tasksToImageStoreRequestsConcurrency,
            taskExternalArchiveTimeouts=settings.taskExternalArchiveTimeouts,
        )

    @staticmethod
    def initializeBaseWorkerClass(settings: SettingsLambdaTasksUnit, monitoring: LunaMonitoringManager | None) -> None:
        """
        Initialize base SubTask worker class.
        Args:
            settings: task worker settings
            monitoring: luna influx monitoring
        """
        BaseSubTask.initFromConfig(
            maxErrorCountPerTask=settings.maxErrorCountPerTask,
            influxMonitoring=monitoring,
            limits=settings.platformLimits,
            facesRequestsConcurrency=settings.tasksToFacesRequestsConcurrency,
            handlersRequestConcurrency=settings.tasksToHandlersConcurrency,
            externalApiAddress=settings.externalApiAddress,
            lisObjectsAddress=settings.objectsOriginStorage,
            lisEnabled=settings.additionalServicesUsage.lunaImageStore,
        )

    @staticmethod
    def initializeMonitoringPoints(serviceName: str):
        """
        Initialize monitoring points classes.
        Args:
            serviceName: custom service name
        """
        WorkerRequestPoint.initialize(serviceName=serviceName)
        SubtaskMonitoringBasePoint.initialize(serviceName=serviceName)
