"""
Module contains settings class for luna-faces
"""

from typing import Optional

from ..settings.classes import (
    DBSetting as _DBSetting,
    RedisKeyStorageSetting as _RedisKeyStorageSetting,
    RedisStorageSetting as _RedisStorageSetting,
    ServiceAddressSettings as _ServiceAddressSettings,
)
from .base_python_service_settings import BasePythonServiceSettingsClass


class SettingsFaces(BasePythonServiceSettingsClass):
    """
    Settings class for luna-faces
    Attributes:
        facesDB(DBSetting): faces database settings
        storageTime(str): storage time setting
        defaultFaceDescriptorVersion(int): default face descriptor version setting
        useMaterialViews(int): material view using flag setting
        databaseNumber(int): database number setting
        dbPingMaxCount (int): max count of database ping tries (<= 0 without ping)
        attributesDB (RedisDBSetting): temporary attributes database settings
        attributesStorageSettings (RedisStorageSetting): temporary attributes storage settings
        licensesAddress(ServiceAddressSettings): luna-licenses address settings
    """

    serviceName = "luna-faces"

    def __init__(self):
        super().__init__()
        self.facesDB: Optional[_DBSetting] = None
        self.storageTime: Optional[str] = None
        self.defaultFaceDescriptorVersion: Optional[int] = None
        self.useMaterialViews: Optional[bool] = None
        self.databaseNumber: Optional[int] = None
        self.dbPingMaxCount: Optional[int] = None
        self.attributesDB: Optional[_RedisStorageSetting] = None
        self.attributesStorageSettings: Optional[_RedisKeyStorageSetting] = None
        self.licensesAddress: Optional[_ServiceAddressSettings] = None

    def applySettings(self):
        """
        Deserialize loaded settings to classes
        """
        super().applySettings()
        self.facesDB = _DBSetting(self.serviceName, self["LUNA_FACES_DB"])

        self.storageTime = self["storage_time"]
        self.defaultFaceDescriptorVersion = self["default_face_descriptor_version"]
        self.useMaterialViews = self["use_material_views"]
        self.databaseNumber = self["database_number"]
        self.dbPingMaxCount = self["luna_faces_db_ping_max_count"]
        self.attributesDB = _RedisStorageSetting(self.serviceName, "LUNA_ATTRIBUTES_DB", self["LUNA_ATTRIBUTES_DB"])
        self.attributesStorageSettings = _RedisKeyStorageSetting(
            self.serviceName, "ATTRIBUTES_STORAGE_POLICY", self["ATTRIBUTES_STORAGE_POLICY"]
        )
        self.licensesAddress = _ServiceAddressSettings(
            self.serviceName, "LUNA_LICENSES_ADDRESS", self["LUNA_LICENSES_ADDRESS"]
        )
