"""
Module contains settings class for luna-tasks
"""

from typing import Optional

from ..settings.classes import (
    AdditionalServicesUsageSetting as _AdditionalServicesUsageSetting,
    DBSetting as _DBSetting,
    ImageStoreAddressSettings as _ImageStoreAddressSettings,
    PlatformLimitsSettings as _PlatformLimitsSettings,
    RedisStorageSetting as _RedisStorageSetting,
    ServiceAddressSettings as _ServiceAddressSettings,
    ServiceTimeoutsSettings as _ServiceTimeoutsSettings,
)
from .base_python_service_settings import BasePythonServiceSettingsClass


class SettingsTasks(BasePythonServiceSettingsClass):
    """
    Settings class for luna-tasks
    Attributes:
        db(DBSetting): database settings
        faceSamplesStorage(ImageStoreAddressSettings): face samples storage settings
        bodySamplesStorage(ImageStoreAddressSettings): body samples storage settings
        imageOriginStorage(ImageStoreAddressSettings): image origin storage settings
        objectsOriginStorage(ImageStoreAddressSettings): objects origin storage settings
        faceSamplesStorageTimeouts(ServiceTimeoutsSettings): Luna Image Store client timeouts for face samples
        bodySamplesStorageTimeouts(ServiceTimeoutsSettings): Luna Image Store client timeouts for body samples
        taskResultStorage(ImageStoreAddressSettings): Luna Image Store address for task results
        taskResultStorageTimeouts(ServiceTimeoutsSettings): Luna Image Store client timeouts for task results
        taskExternalArchiveTimeouts(ServiceTimeoutsSettings): Luna Task load external archive timeouts
        facesAddress(ServiceAddressSettings): Luna Faces address
        facesTimeouts(ServiceTimeoutsSettings): Luna Faces timeouts
        eventsAddress(ServiceAddressSettings): Luna Events address
        eventsTimeouts(ServiceTimeoutsSettings): Luna Events timeouts
        additionalServicesUsage(SupportServicesSetting): support services container
        storageTime(str): time for records in the database
        maxErrorCountPerTask(int): maximum errors count per task to store in the database
        tasksToImageStoreRequestsConcurrency(int): concurrency of requests to Luna Image Store
        tasksToFacesRequestsConcurrency(int): concurrency of requests to Luna Faces
        pythonMatcherAddress(ServiceAddressSettings): luna-python-matcher address settings
        matcherProxyAddress(ServiceAddressSettings): python-matcher-proxy address settings
        tasksToMatcherTimeouts(ServiceTimeoutsSettings): timeouts for luna-python-matcher and matcher-proxy settings
        platformLimits (PlatformLimitsSettings): platform limits
        externalApiAddress (_ServiceAddressSettings): external luna api address
        redisDB (RedisStorageSetting): redis db address
        lambdaTasksUnitTimeouts (_ServiceTimeoutsSettings): luna-lambda-tasks timeouts settings
    """

    serviceName = "luna-tasks"

    def __init__(self):
        super().__init__()
        self.db: Optional[_DBSetting] = None
        self.faceSamplesStorage: Optional[_ImageStoreAddressSettings] = None
        self.bodySamplesStorage: Optional[_ImageStoreAddressSettings] = None
        self.imageOriginStorage: Optional[_ImageStoreAddressSettings] = None
        self.faceSamplesStorageTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.bodySamplesStorageTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.imageOriginStorageTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.taskResultStorage: Optional[_ImageStoreAddressSettings] = None
        self.taskResultStorageTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.taskExternalArchiveTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.facesAddress: Optional[_ServiceAddressSettings] = None
        self.facesTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.eventsAddress: Optional[_ServiceAddressSettings] = None
        self.eventsTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.platformLimits: Optional[_PlatformLimitsSettings] = None

        # LUNA TASKS ROUTER SETTINGS
        self.additionalServicesUsage: Optional[_AdditionalServicesUsageSetting] = None
        self.storageTime: Optional[str] = None

        # LUNA TASKS WORKER SETTINGS
        self.maxErrorCountPerTask: Optional[int] = None
        self.tasksToImageStoreRequestsConcurrency: Optional[int] = None
        self.tasksToFacesRequestsConcurrency: Optional[int] = None
        self.handlersAddress: Optional[_ServiceAddressSettings] = None
        self.handlersTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.tasksToHandlersConcurrency: Optional[int] = None
        self.objectsOriginStorage: Optional[_ImageStoreAddressSettings] = None

        self.pythonMatcherAddress: Optional[_ServiceAddressSettings] = None
        self.matcherProxyAddress: Optional[_ServiceAddressSettings] = None
        self.tasksToMatcherTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.externalApiAddress: Optional[_ServiceAddressSettings] = None
        self.redisDB: Optional[_RedisStorageSetting] = None
        self.lambdaTasksUnitTimeouts: Optional[_ServiceTimeoutsSettings] = None

    def applySettings(self):
        """
        Deserialize loaded settings to classes
        """
        super().applySettings()
        self.db = _DBSetting(self.serviceName, self["LUNA_TASKS_DB"])
        self.faceSamplesStorage = _ImageStoreAddressSettings(
            self.serviceName, "LUNA_IMAGE_STORE_FACES_SAMPLES_ADDRESS", self["LUNA_IMAGE_STORE_FACES_SAMPLES_ADDRESS"]
        )
        self.bodySamplesStorage = _ImageStoreAddressSettings(
            self.serviceName, "LUNA_IMAGE_STORE_BODIES_SAMPLES_ADDRESS", self["LUNA_IMAGE_STORE_BODIES_SAMPLES_ADDRESS"]
        )
        self.imageOriginStorage = _ImageStoreAddressSettings(
            self.serviceName,
            "LUNA_IMAGE_STORE_IMAGES_ADDRESS",
            self["LUNA_IMAGE_STORE_IMAGES_ADDRESS"],
        )
        self.faceSamplesStorageTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_IMAGE_STORE_FACES_SAMPLES_TIMEOUTS", self["LUNA_IMAGE_STORE_FACES_SAMPLES_TIMEOUTS"]
        )
        self.bodySamplesStorageTimeouts = _ServiceTimeoutsSettings(
            self.serviceName,
            "LUNA_IMAGE_STORE_BODIES_SAMPLES_TIMEOUTS",
            self["LUNA_IMAGE_STORE_BODIES_SAMPLES_TIMEOUTS"],
        )
        self.imageOriginStorageTimeouts = _ServiceTimeoutsSettings(
            self.serviceName,
            "LUNA_IMAGE_STORE_IMAGES_TIMEOUTS",
            self["LUNA_IMAGE_STORE_IMAGES_TIMEOUTS"],
        )
        self.taskResultStorage = _ImageStoreAddressSettings(
            self.serviceName, "LUNA_IMAGE_STORE_TASK_RESULT_ADDRESS", self["LUNA_IMAGE_STORE_TASK_RESULT_ADDRESS"]
        )
        self.taskResultStorageTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_IMAGE_STORE_TASK_RESULT_TIMEOUTS", self["LUNA_IMAGE_STORE_TASK_RESULT_TIMEOUTS"]
        )
        self.facesAddress = _ServiceAddressSettings(self.serviceName, "LUNA_FACES_ADDRESS", self["LUNA_FACES_ADDRESS"])
        self.facesTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_FACES_TIMEOUTS", self["LUNA_FACES_TIMEOUTS"]
        )
        self.eventsAddress = _ServiceAddressSettings(
            self.serviceName, "LUNA_EVENTS_ADDRESS", self["LUNA_EVENTS_ADDRESS"]
        )
        self.eventsTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_EVENTS_TIMEOUTS", self["LUNA_EVENTS_TIMEOUTS"]
        )
        self.taskExternalArchiveTimeouts = _ServiceTimeoutsSettings(
            self.serviceName,
            "LUNA_TASKS_LOAD_EXTERNAL_ARCHIVE_TIMEOUTS",
            self["LUNA_TASKS_LOAD_EXTERNAL_ARCHIVE_TIMEOUTS"],
        )
        self.platformLimits = _PlatformLimitsSettings(self.serviceName, self["PLATFORM_LIMITS"])

        # LUNA TASKS ROUTER SETTINGS
        self.additionalServicesUsage = _AdditionalServicesUsageSetting(
            self.serviceName, self["ADDITIONAL_SERVICES_USAGE"]
        )
        self.storageTime = self["storage_time"]

        # LUNA TASKS WORKER SETTINGS
        self.maxErrorCountPerTask = self["max_error_count_per_task"]
        self.tasksToImageStoreRequestsConcurrency = self["tasks_to_image_store_requests_concurrency"]
        self.tasksToFacesRequestsConcurrency = self["tasks_to_faces_requests_concurrency"]
        self.handlersAddress = _ServiceAddressSettings(
            self.serviceName, "LUNA_HANDLERS_ADDRESS", self["LUNA_HANDLERS_ADDRESS"]
        )
        self.handlersTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_HANDLERS_TIMEOUTS", self["LUNA_HANDLERS_TIMEOUTS"]
        )
        self.tasksToHandlersConcurrency = self["TASKS_TO_HANDLERS_CONCURRENCY"]

        self.pythonMatcherAddress = _ServiceAddressSettings(
            self.serviceName, "LUNA_PYTHON_MATCHER_ADDRESS", self["LUNA_PYTHON_MATCHER_ADDRESS"]
        )
        self.matcherProxyAddress = _ServiceAddressSettings(
            self.serviceName, "LUNA_MATCHER_PROXY_ADDRESS", self["LUNA_MATCHER_PROXY_ADDRESS"]
        )
        self.tasksToMatcherTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_TASKS_TO_MATCHER_TIMEOUTS", self["LUNA_TASKS_TO_MATCHER_TIMEOUTS"]
        )
        self.externalApiAddress = _ServiceAddressSettings(
            self.serviceName,
            "EXTERNAL_LUNA_API_ADDRESS",
            self["EXTERNAL_LUNA_API_ADDRESS"],
        )
        self.objectsOriginStorage = _ImageStoreAddressSettings(
            self.serviceName, "LUNA_IMAGE_STORE_OBJECTS_ADDRESS", self["LUNA_IMAGE_STORE_OBJECTS_ADDRESS"]
        )
        self.redisDB = _RedisStorageSetting(self.serviceName, "TASKS_REDIS_DB_ADDRESS", self["TASKS_REDIS_DB_ADDRESS"])
        self.lambdaTasksUnitTimeouts = _ServiceTimeoutsSettings(
            self.serviceName,
            "LUNA_LAMBDA_TASKS_UNIT_TIMEOUTS",
            self["LUNA_LAMBDA_TASKS_UNIT_TIMEOUTS"],
        )
