"""
Module contains tests for current dumps
"""

import unittest
from datetime import date
from pathlib import Path
from typing import Callable, Dict, List, Tuple

from configs_core.migration.classes import OperationMode
from configs_core.migration.migrate import MigrationException, MigrationProcessor

from configs.dump.services import LUNA_PLATFORM

BASE_REVISION_NAME = "base_revision"


class TestRevisionsGetter(unittest.TestCase):
    """Test getting revisions in right order"""

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.revisions = [BASE_REVISION_NAME] + [str(date(year=2020, month=1, day=i)) for i in range(1, 10)]
        cls.sortedRevisions = cls.revisions + ["head"]
        versionsDir: Path = Path(__file__).parent.parent / "configs" / "migration" / "versions"
        testMigrationProcessor = MigrationProcessor(
            "postgres://luna:luna@127.0.0.1:5432/luna_configurator",
            versionsDir,
            services=LUNA_PLATFORM,
            component="platform",
        )
        cls.revisionsGetter: Callable[[Dict[str, str], str, str], Tuple[List[str], OperationMode, str]] = (
            testMigrationProcessor.getRevisionsToMigrate
        )
        cls._revisionToDownRevisionMap = {
            cls.revisions[idx + 1] if idx < (len(cls.revisions) - 1) else "head": rev
            for idx, rev in enumerate(cls.revisions)
        }

    def setUp(self) -> None:
        super().setUp()
        self.revisionToDownRevisionMap = self._revisionToDownRevisionMap.copy()

    def test_get_sorted_revisions(self):
        """Test getting sorted revisions"""
        revisions = MigrationProcessor.getSortedRevisions(self.revisionToDownRevisionMap)
        self.assertListEqual(revisions, self.sortedRevisions)

    def test_get_sorted_revisions_from_empty_revision_map(self):
        """Test getting sorted revisions from empty revision map"""
        revisions = MigrationProcessor.getSortedRevisions({})
        self.assertListEqual(revisions, [BASE_REVISION_NAME])

    def test_get_sorted_revisions_duplicate_heads(self):
        """Test getting sorted revisions with duplicate heads"""
        with self.assertRaises(MigrationException) as exc:
            self.revisionToDownRevisionMap["head2"] = self.sortedRevisions[-2]
            MigrationProcessor.getSortedRevisions(self.revisionToDownRevisionMap)
        self.assertIsInstance(exc.exception, MigrationException)
        self.assertIn("Got duplicated heads revisions", exc.exception.args[0])

    def test_get_revisions_to_migrate_from_base_to_head(self):
        """Test getting migration revisions order from base to head revision"""
        revisions, operationMode, targetRevision = self.revisionsGetter(
            self.revisionToDownRevisionMap, self.sortedRevisions[0], "head"
        )
        self.assertListEqual(revisions, self.sortedRevisions[1::])
        self.assertEqual(operationMode, OperationMode.upgrade)
        self.assertEqual(targetRevision, self.sortedRevisions[-1])

    def test_get_revisions_upgrade_from_any_to_any(self):
        """Test getting migration revisions order from each revision to upper revisions"""
        for curRevIdx, currentRevision in enumerate(self.sortedRevisions):
            if curRevIdx == len(self.sortedRevisions) - 1:
                continue
            for targetRevision in self.sortedRevisions[curRevIdx + 1 :]:
                targetRevIdx = self.sortedRevisions.index(targetRevision)
                with self.subTest(currentRevision=currentRevision, targetRevision=targetRevision):
                    revisionMap = self.revisionToDownRevisionMap.copy()
                    revisions, operationMode, receivedTargetRevision = self.revisionsGetter(
                        revisionMap, currentRevision, targetRevision
                    )
                    self.assertEqual(operationMode, OperationMode.upgrade)
                    self.assertEqual(targetRevision, receivedTargetRevision)
                    self.assertListEqual(revisions, self.sortedRevisions[curRevIdx + 1 : targetRevIdx + 1])

    def test_get_revisions_same_revision(self):
        """Test getting migration revisions order with the same current and target revisions"""
        for revision in self.sortedRevisions:
            with self.subTest(revision=revision):
                revisionMap = self.revisionToDownRevisionMap.copy()
                revisions, _, _ = self.revisionsGetter(revisionMap, revision, revision)
            self.assertListEqual(revisions, [])

    def test_get_revisions_revision_not_found(self):
        """Test getting migration revisions order with non-existing revision"""
        with self.assertRaises(MigrationException) as exc:
            self.revisionsGetter(self.revisionToDownRevisionMap, self.sortedRevisions[0], "revision")
        self.assertEqual("Target revision 'revision' not found", exc.exception.args[0])

    def test_get_revisions_downgrade_from_any_to_any(self):
        """Test getting migration revisions order from each revision to down revisions"""
        for curRevIdx, currentRevision in enumerate(self.sortedRevisions):
            if curRevIdx == 0:
                continue
            for targetRevision in self.sortedRevisions[:curRevIdx]:
                targetRevIdx = self.sortedRevisions.index(targetRevision)
                with self.subTest(currentRevision=currentRevision, targetRevision=targetRevision):
                    revisionMap = self.revisionToDownRevisionMap.copy()
                    revisions, operationMode, receivedTargetRevision = self.revisionsGetter(
                        revisionMap, currentRevision, targetRevision
                    )
                    self.assertEqual(operationMode, OperationMode.downgrade)
                    self.assertEqual(targetRevision, receivedTargetRevision)
                    self.assertListEqual(
                        revisions,
                        self.sortedRevisions[::-1][
                            len(self.sortedRevisions) - curRevIdx - 1 : len(self.sortedRevisions) - targetRevIdx - 1
                        ],
                    )

    def test_get_revisions_downgrade_to_previous(self):
        """Test getting migration revisions order downgrading to previous revision"""
        for curRevIdx, currentRevision in enumerate(self.sortedRevisions):
            if curRevIdx == 0:
                continue
            with self.subTest(currentRevision=currentRevision):
                revisionMap = self.revisionToDownRevisionMap.copy()
                revisions, operationMode, receivedTargetRevision = self.revisionsGetter(
                    revisionMap, currentRevision, "-1"
                )
                self.assertEqual(operationMode, OperationMode.downgrade)
                self.assertEqual(self.sortedRevisions[curRevIdx - 1], receivedTargetRevision)
                self.assertListEqual(
                    revisions,
                    self.sortedRevisions[::-1][
                        len(self.sortedRevisions) - curRevIdx - 1 : len(self.sortedRevisions) - curRevIdx
                    ],
                )
