from argparse import ArgumentParser, ArgumentTypeError
from typing import Literal

from cow.utils.check_connection import (
    checkConnectionToLunaEvents,
    checkConnectionToLunaFaces,
    checkConnectionToLunaImageStore,
    checkConnectionToLunaPythonMatcher,
    checkConnectionToLunaRemoteSDK,
    checkConnectionToLunaSender,
    checkConnectionToRedis,
    checkConnectionToSentinel,
)
from cow.web.checks import checkConnectionToInflux

from .configs.configs.services import SettingsLambdaTasksUnit, SettingsLambdaUnit
from .configs.configs.settings.classes import ImageStoreAddressSettings, ServiceAddressSettings, ServiceTimeoutsSettings


def positiveInt(value):
    """Positive integer getter"""
    ivalue = int(value)
    if ivalue <= 0:
        raise ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue


def getParsedCMDArgs(lambdaType: Literal["handlers", "tasks"] = "handlers"):
    """Get parsed cmd args"""
    description = "Lambda application" if lambdaType == "handlers" else "Lambda tasks application"
    name = "lambda-unique-name" if lambdaType == "handlers" else "lambda-task-unique-name"
    nameHelp = "unique lambda name" if lambdaType == "handlers" else "unique lambda tasks name"
    parser = ArgumentParser(description=description, add_help=False)
    parser.add_argument("--name", default=name, type=str, help=nameHelp)
    parser.add_argument("--version", default=1, type=positiveInt, help="lambda version")
    if lambdaType == "handlers":
        parser.add_argument("--lambda-type", default="handlers", choices=["handlers", "standalone"], help="lambda type")
    parser.add_argument("--config", type=str, help="path to config")
    parser.add_argument(
        "--config-reload",
        "-cf",
        default=0,
        type=int,
        help="reload configuration in real time",
        const=True,
        nargs="?",
        choices=[0, 1],
    )
    parser.add_argument(
        "--luna-config",
        default=None,
        type=str,
        metavar="=",
        help="origin and api version of luna-configurator service for pulling setting from it "
        '(e.g. "http://127.0.0.1:5070/1")',
    )

    cmdArgs, _ = parser.parse_known_args()
    return cmdArgs


def getChecks4StandaloneLambda(config: SettingsLambdaUnit):
    """Check for standalone lambda"""
    checks = []
    if config.monitoring.sendData:
        checks.append(checkConnectionToInflux(config))
    return checks


def getChecks4TasksLambda(config: SettingsLambdaTasksUnit):
    """Check for tasks lambda"""
    checks = []
    if config.additionalServicesUsage.lunaImageStore:
        checks.append(
            checkConnectionToLunaImageStore(
                postfix="task_results",
                **getLISClientKwargs(
                    addressSetting=config.taskResultStorage, timeoutSettings=config.taskResultStorageTimeouts
                ),
            )
        )
    if config.redisDB.sentinel.sentinels:
        checks.append(
            checkConnectionToSentinel(
                sentinels=config.redisDB.sentinel.sentinels,
                masterName=config.redisDB.sentinel.masterName,
                redisPassword=config.redisDB.password,
                asyncCheck=True,
            )
        )
    else:
        checks.append(
            checkConnectionToRedis(
                redisHost=config.redisDB.host,
                redisPort=config.redisDB.port,
                redisPassword=config.redisDB.password,
                asyncCheck=True,
            )
        )
    if config.monitoring.sendData:
        checks.append(checkConnectionToInflux(config))
    return checks


def getServiceClientKwargs(addressSetting: ServiceAddressSettings, timeoutSettings: ServiceTimeoutsSettings) -> dict:
    """Get handlers client kwargs"""
    return {
        "origin": addressSetting.origin,
        "apiVersion": addressSetting.apiVersion,
        "totalTimeout": timeoutSettings.totalTimeout,
        "connectTimeout": timeoutSettings.connectTimeout,
        "sockConnectTimeout": timeoutSettings.sockConnectTimeout,
        "sockReadTimeout": timeoutSettings.sockReadTimeout,
        "asyncCheck": True,
    }


def getLISClientKwargs(addressSetting: ImageStoreAddressSettings, timeoutSettings: ServiceTimeoutsSettings) -> dict:
    """Get LIS client kwargs"""
    return {
        "origin": addressSetting.origin,
        "apiVersion": addressSetting.apiVersion,
        "bucket": addressSetting.bucket,
        "totalTimeout": timeoutSettings.totalTimeout,
        "connectTimeout": timeoutSettings.connectTimeout,
        "sockReadTimeout": timeoutSettings.sockReadTimeout,
        "sockConnectTimeout": timeoutSettings.sockConnectTimeout,
        "asyncCheck": True,
    }


def getChecks4HandlersLambda(config: SettingsLambdaUnit):
    """Check for handlers lambda"""
    checks = getChecks4StandaloneLambda(config=config)
    checks.extend(
        [
            checkConnectionToLunaFaces(
                **getServiceClientKwargs(addressSetting=config.facesAddress, timeoutSettings=config.facesTimeouts),
            ),
            checkConnectionToLunaRemoteSDK(
                **getServiceClientKwargs(
                    addressSetting=config.remoteSDKAddress, timeoutSettings=config.remoteSDKTimeouts
                ),
            ),
        ]
    )
    if config.additionalServicesUsage.lunaImageStore:
        checks.extend(
            [
                checkConnectionToLunaImageStore(
                    postfix="faces_samples",
                    **getLISClientKwargs(
                        addressSetting=config.faceSamplesStorage, timeoutSettings=config.faceSamplesStorageTimeouts
                    ),
                ),
                checkConnectionToLunaImageStore(
                    postfix="bodies_samples",
                    **getLISClientKwargs(
                        addressSetting=config.bodySamplesStorage, timeoutSettings=config.bodySamplesStorageTimeouts
                    ),
                ),
                checkConnectionToLunaImageStore(
                    postfix="images",
                    **getLISClientKwargs(
                        addressSetting=config.imageOriginStorage, timeoutSettings=config.imageOriginStorageTimeouts
                    ),
                ),
            ]
        )
        if config.additionalServicesUsage.lunaEvents:
            checks.append(
                checkConnectionToLunaEvents(
                    **getServiceClientKwargs(
                        addressSetting=config.eventsAddress, timeoutSettings=config.eventsTimeouts
                    ),
                )
            )
        if config.additionalServicesUsage.lunaSender:
            checks.append(
                checkConnectionToLunaSender(
                    **getServiceClientKwargs(
                        addressSetting=config.senderAddress,
                        timeoutSettings=ServiceTimeoutsSettings(
                            "luna-sender",
                            "luna-sender",
                            {
                                "connect": 60,
                                "request": 60,
                                "sock_connect": 60,
                                "sock_read": 60,
                            },
                        ),
                    )
                )
            )
        if config.additionalServicesUsage.lunaMatcherProxy:
            checks.append(
                checkConnectionToLunaPythonMatcher(
                    **getServiceClientKwargs(
                        addressSetting=config.matcherProxyAddress, timeoutSettings=config.matcherProxyTimeouts
                    ),
                )
            )
        else:
            checks.append(
                checkConnectionToLunaPythonMatcher(
                    **getServiceClientKwargs(
                        addressSetting=config.pythonMatcherAddress, timeoutSettings=config.pythonMatcherTimeouts
                    ),
                )
            )
    return checks
