import abc
import asyncio
import inspect
import os
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from time import time
from zipfile import ZipFile

import ujson
from cryptography.fernet import InvalidToken
from luna3.client import Client
from luna3.common.exceptions import LunaApiException
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import makeRequest
from luna3.image_store.image_store import StoreApi
from vlutils.descriptors.data import DescriptorType

from .enums import SourceType

projectName = Path(__file__).parent.parent.stem

if projectName == "luna_tasks":
    from configs.configs.configs.settings.classes import (
        ImageStoreAddressSettings,
        PlatformLimitsSettings,
        ServiceAddressSettings,
        ServiceTimeoutsSettings,
    )
    from crutches_on_wheels.cow.enums.tasks import SubTaskStatus, TaskStatus, TaskType
    from crutches_on_wheels.cow.errors.errors import Error, ErrorInfo
    from crutches_on_wheels.cow.errors.exception import VLException
    from crutches_on_wheels.cow.monitoring.manager import LunaMonitoringManager
    from crutches_on_wheels.cow.utils.initializable import AsyncRunner
    from crutches_on_wheels.cow.utils.log import Logger
    from crutches_on_wheels.cow.web.error_handlers import createErrorFromLunaApiException
else:
    from ..configs.configs.settings.classes import (
        ImageStoreAddressSettings,
        PlatformLimitsSettings,
        ServiceAddressSettings,
        ServiceTimeoutsSettings,
    )
    from cow.enums.tasks import SubTaskStatus, TaskStatus, TaskType
    from cow.errors.errors import Error, ErrorInfo
    from cow.errors.exception import VLException
    from cow.monitoring.manager import LunaMonitoringManager
    from cow.utils.initializable import AsyncRunner
    from cow.utils.log import Logger
    from cow.web.error_handlers import createErrorFromLunaApiException

from .monitoring.points import SubtaskErrorPoint, SubtaskExecutionPoint, SubtaskFinalPoint, WorkerQueuePoint
from .redis_db import structs
from .redis_db.context import WorkerRedisContext

_APPROXIMATE_LAN_SPEED = 10_000_000  # bytes/s
_REQUEST_BODY_LENGTH = _APPROXIMATE_LAN_SPEED * 60  # default request timeout - 60 sec


@dataclass
class SubTaskMonitoringData:
    """
    Structure for saving an additional monitoring data.
    """

    # tags: tags
    tags: dict[str, int | float | str] = field(default_factory=dict)
    # fields: fields
    fields: dict[str, int | float | str] = field(default_factory=dict)

    def addData(self, tags: dict | None = None, fields: dict | None = None) -> None:
        """
        Add data to monitoring
        Args:
            tags: new tags
            fields:  new fields
        """
        if tags is not None:
            self.tags.update(tags)
        if fields is not None:
            self.fields.update(fields)


class SubTaskToImageStore:
    """
    Class for work with luna-image-store service.

    Attributes:
        luna3Client: luna3 client with the request id
        logger: service logger
    """

    # task result storage bucket
    resultStoreBucket: str
    # face sample bucket
    faceSampleBucket: str
    # body sample bucket
    bodySampleBucket: str
    # image bucket
    imageOriginBucket: str
    # task result storage timeouts
    taskResultStorageTimeouts: ServiceTimeoutsSettings
    # face sample storage timeouts
    faceSamplesStorageTimeouts: ServiceTimeoutsSettings
    # body sample storage timeouts
    bodySamplesStorageTimeouts: ServiceTimeoutsSettings
    # image origin storage timeouts
    imageOriginStorageTimeouts: ServiceTimeoutsSettings
    # task load external archive timeouts
    taskExternalArchiveTimeouts: ServiceTimeoutsSettings
    # task to image store request concurrency
    imageStoreRequestConcurrency: int

    def __init__(self, luna3Client: Client, logger: Logger):
        self.luna3Client = luna3Client
        self.logger = logger

    @classmethod
    def initFromConfig(
        cls,
        resultStoreBucket: str,
        faceSampleBucket: str,
        bodySampleBucket: str,
        imageOriginBucket: str,
        taskResultStorageTimeouts: ServiceTimeoutsSettings,
        faceSamplesStorageTimeouts: ServiceTimeoutsSettings,
        bodySamplesStorageTimeouts: ServiceTimeoutsSettings,
        imageOriginStorageTimeouts: ServiceTimeoutsSettings,
        imageStoreRequestConcurrency: int,
        taskExternalArchiveTimeouts: ServiceTimeoutsSettings,
    ):
        """
        Init from config. Must be called at application startup
        Args:
            resultStoreBucket: task result storage bucket
            faceSampleBucket: face sample bucket
            bodySampleBucket: body sample bucket
            imageOriginBucket: image origin bucket
            taskResultStorageTimeouts: task result storage timeouts
            faceSamplesStorageTimeouts: face sample storage timeouts
            bodySamplesStorageTimeouts: body sample storage timeouts
            imageOriginStorageTimeouts: image origin storage timeouts
            imageStoreRequestConcurrency: task to image store request concurrency
            taskExternalArchiveTimeouts: task load external archive timeouts
        """
        cls.resultStoreBucket = resultStoreBucket
        cls.faceSampleBucket = faceSampleBucket
        cls.bodySampleBucket = bodySampleBucket
        cls.imageOriginBucket = imageOriginBucket
        cls.taskResultStorageTimeouts = taskResultStorageTimeouts
        cls.faceSamplesStorageTimeouts = faceSamplesStorageTimeouts
        cls.bodySamplesStorageTimeouts = bodySamplesStorageTimeouts
        cls.imageOriginStorageTimeouts = imageOriginStorageTimeouts
        cls.imageStoreRequestConcurrency = imageStoreRequestConcurrency
        cls.taskExternalArchiveTimeouts = taskExternalArchiveTimeouts

    def getTotalRequestTimeout(self, body: str | bytes) -> int:
        """
        Calculate request timeout.

        Args:
            body: request body
        Returns:
            request timeout
        """
        bodyBytes = body if isinstance(body, bytes) else body.encode("utf-8")
        bodyLength = len(bodyBytes)

        if _REQUEST_BODY_LENGTH < bodyLength:
            return round(bodyLength / _REQUEST_BODY_LENGTH * self.taskResultStorageTimeouts.totalTimeout)
        return self.taskResultStorageTimeouts.totalTimeout

    async def postTaskResult(
        self, objectBody: str, accountId: str, calculateTotalTimeouts: bool = False, **kwargs
    ) -> LunaResponse:
        """
        Create object to task result bucket.

        Args:
            objectBody: object, available: text, json
            accountId: luna account id
            calculateTotalTimeouts: calculating the total request timeout depending on request body size, if `True`
        Returns:
            luna response
        """
        if calculateTotalTimeouts:
            kwargs["totalTimeout"] = self.getTotalRequestTimeout(objectBody)
        return await self.luna3Client.lunaTasksResultsStore.postObject(
            objectBody=objectBody, bucketName=self.resultStoreBucket, accountId=accountId, **kwargs
        )

    async def getTaskResult(self, objectId: str, **kwargs) -> LunaResponse:
        """
        Get task result object from store.

        Args:
            objectId: result object id
        Returns:
            luna response
        """
        return await self.luna3Client.lunaTasksResultsStore.getObject(
            bucketName=self.resultStoreBucket, objectId=objectId, **kwargs
        )

    async def loadImageToZip(
        self,
        targetZipFile: ZipFile,
        objectIdToSourceMap: dict[str, str],
        sourceType: SourceType,
        descriptorType: DescriptorType,
        accountId: str,
        taskId: int,
    ) -> None:
        """
        Load images from self.objectIdToUrlMap to self.targetZipFile ZipFile.
        """

        async def loadAndSaveOne(
            objectId: str, sampleBucket: str, storeClient: StoreApi, serviceTimeouts: ServiceTimeoutsSettings
        ) -> None:
            """
            Load and save image corresponding to the object.
            if self.sourceType is sample    - load sample from image store
            if source (url) starts with "/" - get sample id from source (relative url) and load sample from image store
            otherwise                       - load image by provided source (full url)

            Args:
                objectId: object id to load image for
                sampleBucket: bucket for image store
                storeClient: image store client
                serviceTimeouts: image store timeouts settings
            """
            source = objectIdToSourceMap[objectId]
            if sourceType == SourceType.sample:
                # sample: "{uuid}"
                reply = await storeClient.getImage(sampleBucket, source, accountId=accountId)
            elif source.startswith("/"):
                # sample: "/1/samples/{uuid}"
                sampleId = source.split("/")[-1]
                reply = await storeClient.getImage(sampleBucket, sampleId, accountId=accountId)
            else:
                # url: "http://www.example.com/1.jpg"
                # field extensibility https://tools.ietf.org/html/rfc7230#section-3.2.1
                kwargs = dict(
                    totalTimeout=serviceTimeouts.totalTimeout,
                    connectTimeout=serviceTimeouts.connectTimeout,
                    sockConnectTimeout=serviceTimeouts.sockConnectTimeout,
                    sockReadTimeout=serviceTimeouts.sockReadTimeout,
                    headers={"Luna-Request-Id": storeClient.getRequestId(), "Luna-Account-Id": accountId},
                    asyncRequest=True,
                )
                if f"buckets/{sampleBucket}/images" in source:
                    # detect that source is a samples bucket, add account id, fix avatar injections
                    queryParams = {"account_id": accountId}
                else:
                    queryParams = None
                reply = await makeRequest(source, "GET", queryParams=queryParams, **kwargs)

            if reply.success:
                fullImageName = os.path.join(str(taskId), "images", f"{objectId}.jpg")
                targetZipFile.writestr(fullImageName, reply.body)
            else:
                self.logger.warning(
                    f"Unexpected response on loading image of the {sourceType.value[:-1]} "
                    f'"{objectId}": {reply.statusCode} "{reply.text}"'
                )

        if descriptorType == DescriptorType.face:
            saveObject = partial(
                loadAndSaveOne,
                sampleBucket=self.faceSampleBucket,
                storeClient=self.luna3Client.lunaFaceSamplesStore,
                serviceTimeouts=self.faceSamplesStorageTimeouts,
            )
        else:
            saveObject = partial(
                loadAndSaveOne,
                sampleBucket=self.bodySampleBucket,
                storeClient=self.luna3Client.lunaBodySamplesStore,
                serviceTimeouts=self.bodySamplesStorageTimeouts,
            )

        coros = map(saveObject, objectIdToSourceMap)
        await AsyncRunner(self.imageStoreRequestConcurrency).runWait(coros, close=True)


class SubTaskLogger(Logger):
    """Sub task logger"""

    def updateRequestId(self, requestId: str):
        """Update request id"""
        self._requestId = requestId

    def resetRequestId(self):
        """Reset request id"""
        self._requestId = self.defaultRequestId


class BaseSubTask(abc.ABC):
    """
    Base sub task class
    Attributes:
        _redisContext: redis context
        _task: current running asyncio task
        _separationTask: current separation task
        _taskId: current processing task id
        _subTaskId: current processing sub task id
        logger: logger
        subTaskToImageStore: sub task to image store processor
        _data: current task data
        _isRunning: running flag
    """

    def __init__(self, redisContext: WorkerRedisContext, luna3Client, loggerTitle: str):
        self.client = luna3Client
        self._redisContext = redisContext
        self._task: asyncio.Task | None = None
        self._separationTask: asyncio.Task | None = None
        self._taskId: bytes | None = None
        self._subTaskId: bytes | None = None
        self.logger = SubTaskLogger(loggerTitle)
        self.subTaskToImageStore = SubTaskToImageStore(luna3Client, self.logger)
        self._data: structs.TaskContentRequest | structs.QueueSubTask | structs.ResultAggregatorSubTask | None = None
        self._isRunning = True

    async def onShutDown(self):
        """
        Perform shutdown actions
            - cancel current task
            - return task data to redis
        """
        self._isRunning = False
        if self._task is not None and not self._task.done():
            await self._redisContext.sendToRedis(
                structs.SubTaskCancel(
                    taskId=self._data.taskId,
                    subTaskId=getattr(self._data, "subTaskId", None),
                    lRId=self._data.lRId,
                    taskType=self._data.taskType,
                )
            )
            self._task.cancel()
            self.logger.info("Current task has been failed due to worker shutdown")
        if self._separationTask is not None and not self._separationTask.done():
            self._separationTask.cancel()
            self.logger.info("Current content separation failed due to worker shutdown")

    @property
    @abc.abstractmethod
    def taskType(self) -> TaskType:
        """Task type"""

    @classmethod
    def initFromConfig(
        cls,
        influxMonitoring: LunaMonitoringManager,
        maxErrorCountPerTask: int,
        limits: PlatformLimitsSettings,
        facesRequestsConcurrency: int,
        handlersRequestConcurrency: int,
        externalApiAddress: ServiceAddressSettings,
        lisObjectsAddress: ImageStoreAddressSettings,
        lisEnabled: bool,
    ) -> None:
        """
        Init from config. Must be called at application startup
        Args:
            maxErrorCountPerTask: max error count per tasks
            influxMonitoring: luna influx monitoring
            limits: platform limits
            facesRequestsConcurrency: task to faces request concurrency
            handlersRequestConcurrency: task to handler request concurrency
            externalApiAddress: external luna-api service address
            lisObjectsAddress: LIS objects address
            lisEnabled: whether lis is enabled
        """
        cls.influxMonitoring = influxMonitoring
        cls.maxErrorCountPerTask = maxErrorCountPerTask
        cls.limits = limits
        cls.facesRequestsConcurrency = facesRequestsConcurrency
        cls.handlersRequestConcurrency = handlersRequestConcurrency
        cls.externalApiAddress = externalApiAddress
        cls.lisObjectsAddress = lisObjectsAddress
        cls.lisEnabled = lisEnabled

    def _resetParams(self):
        """Reset data and current task params"""
        self._data = None
        self._task = None
        self._taskId = None
        self._subTaskId = None

    async def separationStart(self, lambdaId: str | None = None):
        """Start separation processing"""
        while self._isRunning:
            await asyncio.sleep(0.05)
            try:
                if self._separationTask is not None and self._separationTask.done():
                    self._separationTask = None
                elif taskContentRequest := (
                    await self._redisContext.getSeparationRequest(taskType=self.taskType, lambdaId=lambdaId)
                ):
                    self._separationTask = asyncio.create_task(self._splitTask(taskContentRequest=taskContentRequest))
            except Exception:
                self.logger.exception()
                self._redisContext.reInitClient()

    async def start(self, lambdaId: str | None = None):
        """Start task processing"""
        while self._isRunning:
            await asyncio.sleep(0.05)
            try:
                if self._task is not None:
                    if self._task.done():
                        self._resetParams()
                    else:
                        await self._canceller()
                else:
                    if aggregatorTaskData := (
                        await self._redisContext.getResultAggregatorSubTask(taskType=self.taskType, lambdaId=lambdaId)
                    ):
                        self._data = aggregatorTaskData
                        self._taskId = str(aggregatorTaskData.taskId).encode()
                        self._task = asyncio.create_task(self._collectResults(aggregatorTaskData))
                    elif taskData := (
                        await self._redisContext.getSubTask2Process(taskType=self.taskType, lambdaId=lambdaId)
                    ):
                        self._data = taskData
                        self._subTaskId = str(taskData.subTaskId).encode()
                        self._task = asyncio.create_task(self._processTask(taskData))
            except Exception:
                self.logger.exception()
                self._redisContext.reInitClient()

    async def _splitTask(self, taskContentRequest: structs.TaskContentRequest):
        """Split task content to sub task contents"""
        self.logger.updateRequestId(taskContentRequest.lRId)
        self.logger.info(f"Start {taskContentRequest.taskType} task separation")
        contents = await self.splitTask2SubTasks(taskContentRequest)
        await self._redisContext.sendToRedis(
            structs.SubTasksContentReply(
                redisTaskId=taskContentRequest.redisTaskId,
                taskType=self.taskType,
                contents=contents,
                lRId=taskContentRequest.lRId,
            )
        )
        self.logger.info(f"Finish {taskContentRequest.taskType} task separation")

    async def _canceller(self):
        """Check whether current task is need to be cancelled"""
        if await self._redisContext.isNeedToCancel(
            taskType=self.taskType, subTaskId=self._subTaskId, taskId=self._taskId
        ):
            self.logger.resetRequestId()
            isCollectionResultRunning = self._taskId is not None
            if isCollectionResultRunning:
                self.logger.info(f"Collection results for task with id {self._taskId.decode()} has been cancelled")
            else:
                self.logger.info(f"Subtask with id {self._subTaskId.decode()} has been cancelled")
            self._task.cancel()
            self._resetParams()

    @abc.abstractmethod
    async def execute(self, taskData: structs.QueueSubTask) -> tuple[str, SubTaskStatus]:
        """Execute current sub task processing"""
        ...

    @staticmethod
    def getKwargsFromTaskData(
        taskData: structs.QueueSubTask | structs.ResultAggregatorSubTask | structs.SubTaskError | structs.SubTaskResult,
    ):
        """
        Get kwargs from task data
        """
        return {
            "taskId": taskData.taskId,
            "taskType": taskData.taskType,
            "subTaskId": taskData.subTaskId,
            "lRId": taskData.lRId,
        }

    async def _processTask(self, taskData: structs.QueueSubTask):
        """Process task"""
        startTime = time()
        self.influxMonitoring.flushPoints(
            (
                WorkerQueuePoint(
                    time(),
                    taskData.subTaskId,
                    taskId=taskData.taskId,
                    transportTime=time() - taskData.startTime,
                    taskType=taskData.taskType,
                ),
            )
        )

        self.logger.updateRequestId(taskData.lRId)
        self.logger.info(f"Start processing subtask with id {taskData.subTaskId} for task with id {taskData.taskId}")
        await self._redisContext.sendToRedis(structs.SubTaskStart(**self.getKwargsFromTaskData(taskData=taskData)))
        try:
            resultId, status = await self.execute(taskData)
        except Exception as exception:
            resultId, status = None, SubTaskStatus.failed
            if isinstance(exception, VLException):
                await self.createSubTaskError(
                    structs.SubTaskError(error=exception.error, **self.getKwargsFromTaskData(taskData=taskData))
                )
            elif isinstance(exception, LunaApiException):
                await self.createSubTaskError(
                    structs.SubTaskError(
                        error=ErrorInfo.fromDict(exception.json), **self.getKwargsFromTaskData(taskData=taskData)
                    )
                )
            elif taskData.taskType == TaskType.lambda_task:
                await self.createSubTaskError(
                    structs.SubTaskError(
                        error=Error.LambdaError.format(f"{type(exception)}: {str(exception)}"),
                        **self.getKwargsFromTaskData(taskData=taskData),
                    )
                )
            elif isinstance(exception, InvalidToken):
                await self.createSubTaskError(
                    structs.SubTaskError(error=Error.EncryptionError, **self.getKwargsFromTaskData(taskData=taskData))
                )
            else:
                await self.createSubTaskError(
                    structs.SubTaskError(error=Error.UnknownError, **self.getKwargsFromTaskData(taskData=taskData))
                )

        await self._redisContext.sendToRedis(
            structs.SubTaskResult(
                resultId=resultId,
                status=status,
                accountId=taskData.accountId,
                **self.getKwargsFromTaskData(taskData=taskData),
            )
        )
        self.logger.info(f"Finish processing subtask with id {taskData.subTaskId} for task with id {taskData.taskId}")
        self.influxMonitoring.flushPoints(
            (
                SubtaskFinalPoint(
                    time(),
                    subTaskId=taskData.subTaskId,
                    taskId=taskData.taskId,
                    status=status,
                    taskType=taskData.taskType,
                    executionTime=time() - startTime,
                ),
            )
        )

    async def createSubTaskError(self, subTaskError: structs.SubTaskError):
        """Create sub task error"""
        self.logger.updateRequestId(subTaskError.lRId)
        await self._redisContext.sendToRedis(
            structs.SubTaskError(
                error=subTaskError.error,
                additionalInfo=subTaskError.additionalInfo,
                **self.getKwargsFromTaskData(taskData=subTaskError),
            )
        )
        self.logger.info(
            f"An error occurred during subtask with id {subTaskError.subTaskId} for task with id {subTaskError.taskId}"
        )
        self.influxMonitoring.flushPoints(
            (
                SubtaskErrorPoint(
                    time(),
                    taskId=subTaskError.taskId,
                    subtaskId=subTaskError.subTaskId,
                    taskType=subTaskError.taskType,
                    errorCode=subTaskError.error.errorCode,
                ),
            )
        )

    @abc.abstractmethod
    async def splitTask2SubTasks(self, taskContentRequest: structs.TaskContentRequest) -> list[dict]:
        """Split task content to sub task contents"""
        ...

    def exceptionHandler(self, asyncRunner_: AsyncRunner, context: dict):
        """
        Handle exceptions.

        Args:
            asyncRunner_: a AsyncRunner
            context: having the following format:
                    - message: error message, str
                    - job: failed job, aiojobs._job.Job instance
                    - exception: caught exception, Exception instance):
        """
        exc = context["exception"]
        if not isinstance(exc, VLException) or exc.isCriticalError:
            self.logger.resetRequestId()
            self.logger.error(f"an exception occurred: {exc!r}")

    async def _collectResultsFromSingleSubTask(self, resultAggregatorTask: structs.ResultAggregatorSubTask) -> None:
        """
        Collect results of subtask(s), put them to luna-image-store-objects and finish task

        Returns:
            whether the task was successfully finished
        """
        kwargs = dict(
            taskId=resultAggregatorTask.taskId,
            taskType=resultAggregatorTask.taskType,
            lRId=resultAggregatorTask.lRId,
        )
        try:
            self.logger.updateRequestId(resultAggregatorTask.lRId)
            self.logger.info(f"Result collection for task {resultAggregatorTask.taskId} started")
            subTaskResult = resultAggregatorTask.subTaskResults[0]

            errorCount = len(resultAggregatorTask.taskErrors)
            if errorCount > self.maxErrorCountPerTask:
                self.logger.warning(
                    f"Error count {errorCount} is exceeded max allowed number, "
                    f"will be saved only {self.maxErrorCountPerTask} errors in result"
                )
            if errorCount:
                taskResult = {"errors": resultAggregatorTask.taskErrors}
                resultId = (
                    await self.subTaskToImageStore.postTaskResult(
                        objectBody=ujson.dumps(taskResult),
                        accountId=resultAggregatorTask.accountId,
                        calculateTotalTimeouts=True,
                        ttl=resultAggregatorTask.resultStoragePolicy.get("ttl"),
                        raiseError=True,
                    )
                ).json["object_id"]
            else:
                resultId = subTaskResult.resultId

            await self._redisContext.sendToRedis(
                structs.TaskResult(status=TaskStatus.done, resultId=resultId, **kwargs)
            )
            self.logger.info(f"Result collection for task {resultAggregatorTask.taskId} finished")
        except Exception as exc:
            self.logger.error(
                f"Exception occurred during task {resultAggregatorTask.taskId} results collection: {str(exc)}"
            )
            self.logger.error("Exception details:", exc_info=True)
            await self._redisContext.sendToRedis(structs.TaskResult(status=TaskStatus.failed, resultId=None, **kwargs))

    async def _collectResults(self, resultAggregatorTask: structs.ResultAggregatorSubTask) -> None:
        """
        Collect results of subtask(s), put them to luna-image-store-objects and finish task

        Returns:
            whether the task was successfully finished
        """
        kwargs = dict(
            taskId=resultAggregatorTask.taskId,
            taskType=resultAggregatorTask.taskType,
            lRId=resultAggregatorTask.lRId,
        )
        try:
            self.logger.updateRequestId(resultAggregatorTask.lRId)
            self.logger.info(f"Result collection for task {resultAggregatorTask.taskId} started")
            taskResult = {"errors": resultAggregatorTask.taskErrors}

            async def collectTaskResultFromSubTaskResult(subTaskResult: structs.SubTaskResult) -> None:
                """
                Collect task result from subtask result
                """
                isSubTaskFailed = subTaskResult.status == SubTaskStatus.failed

                if self.lisEnabled and not isSubTaskFailed:
                    replyResult = await self.subTaskToImageStore.getTaskResult(objectId=subTaskResult.resultId)
                    if not replyResult.success:
                        self.logger.error(
                            f"Cannot get subtask result '{subTaskResult.resultId}': {replyResult.statusCode} '{replyResult.text}'"
                        )
                        await self.createSubTaskError(
                            subTaskError=structs.SubTaskError(
                                error=ErrorInfo.fromDict(replyResult.json),
                                **self.getKwargsFromTaskData(taskData=subTaskResult),
                            ),
                        )
                        return

                    result = replyResult.json
                    if isinstance(result, list):
                        taskResult.setdefault("result", []).extend(result)
                    elif isinstance(result, dict):
                        if "result" not in taskResult:
                            taskResult["result"] = result
                            return
                        for key, value in result.items():
                            if isinstance(value, list):
                                taskResult["result"][key].extend(value)
                            else:
                                taskResult["result"][key] = value
                    else:
                        raise ValueError(f"Got wrong subtask result type: {type(result)}: {result}")

            if self.lisEnabled and resultAggregatorTask.storeResults:
                asyncRunner = AsyncRunner(
                    self.subTaskToImageStore.imageStoreRequestConcurrency, exceptionHandler=self.exceptionHandler
                )
                await asyncRunner.runWait(
                    (collectTaskResultFromSubTaskResult(row) for row in resultAggregatorTask.subTaskResults)
                )

            if self.lisEnabled:
                response = await self.subTaskToImageStore.postTaskResult(
                    objectBody=ujson.dumps(taskResult),
                    accountId=resultAggregatorTask.accountId,
                    calculateTotalTimeouts=True,
                    ttl=resultAggregatorTask.resultStoragePolicy.get("ttl"),
                )
                if not response.success:
                    self.logger.error(f"Cannot put task result: {response.statusCode} '{response.text}'")
                    await self.createSubTaskError(
                        structs.SubTaskError(
                            error=ErrorInfo.fromDict(response.json),
                            **self.getKwargsFromTaskData(taskData=resultAggregatorTask),
                        )
                    )

                resultId = response.json["object_id"]
            else:
                resultId = None

            await self._redisContext.sendToRedis(
                structs.TaskResult(status=TaskStatus.done, resultId=resultId, **kwargs)
            )
            self.logger.info(f"Result collection for task {resultAggregatorTask.taskId} finished")
        except Exception as exc:
            self.logger.error(
                f"Exception occurred during task {resultAggregatorTask.taskId} results collection: {str(exc)}"
            )
            self.logger.error("Exception details:", exc_info=True)
            await self._redisContext.sendToRedis(structs.TaskResult(status=TaskStatus.failed, resultId=None, **kwargs))

    def flushExecutionMonitoringData(
        self,
        taskData: structs.QueueSubTask,
        fields: dict | None = None,
        tags: dict | None = None,
    ) -> None:
        """
        Flush a data about an execution subtask to the monitoring.

        Args:
            taskData: task data
            fields: fields
            tags: tags
        """
        self.influxMonitoring.flushPoints(
            (
                SubtaskExecutionPoint(
                    time(),
                    subtaskId=taskData.subTaskId,
                    taskId=taskData.taskId,
                    taskType=taskData.taskType,
                    additionalTags=tags,
                    additionalFields=fields,
                ),
            )
        )

    async def saveResultToImageStore(
        self,
        result: bytes | dict | list,
        taskData: structs.QueueSubTask,
        contentType="application/json",
        totalTimeout: int = None,
        sockReadTimeout: int = None,
    ) -> str | None:
        """
        Save a task result to image store

        Args:
            taskData: task data
            result: the result
            contentType: content type.
            totalTimeout: total timeout
            sockReadTimeout: sock read timeout
        Returns:
            object_id from image store
        """
        if not self.lisEnabled or not taskData.content.get("store_results", True):
            return

        startSaveResultTime = time()
        if contentType == "application/json" and not inspect.isasyncgen(result):
            body = ujson.dumps(result)
        else:
            body = result
        try:
            requestKwargs = dict(
                objectBody=body,
                contentType=contentType,
                accountId=taskData.accountId,
                ttl=taskData.resultStoragePolicy.get("ttl"),
            )
            if totalTimeout is not None:
                requestKwargs["totalTimeout"] = totalTimeout
            if sockReadTimeout is not None:
                requestKwargs["sockReadTimeout"] = sockReadTimeout
            response = await self.subTaskToImageStore.postTaskResult(**requestKwargs)
            if not response.success:
                self.logger.updateRequestId(taskData.lRId)
                self.logger.error(f"Cannot put subtask result: {response.statusCode} '{response.text}'")
                await self.createSubTaskError(
                    subTaskError=structs.SubTaskError(
                        error=createErrorFromLunaApiException(response), **self.getKwargsFromTaskData(taskData=taskData)
                    )
                )
                return
            return response.json["object_id"]
        finally:
            self.flushExecutionMonitoringData(
                fields={"save_result_time": time() - startSaveResultTime},
                taskData=taskData,
            )

    async def createErrors(self, taskId: int, subTaskId: int, taskType: TaskType, errors: list, lRId: str):
        errorRequests = []
        for error in errors:
            if "reference_id" not in error:
                continue
            self.logger.updateRequestId(lRId)
            self.logger.error(f"Error in match. Reference: '{error['reference_id']}'. Error: '{error['error']}'.")
            errorRequests.append(
                self.createSubTaskError(
                    structs.SubTaskError(
                        taskId=taskId,
                        subTaskId=subTaskId,
                        taskType=taskType,
                        error=ErrorInfo.fromDict(error["error"]),
                        additionalInfo=error["reference_id"],
                        lRId=lRId,
                    )
                )
            )
        await asyncio.gather(*errorRequests)
